"""Contains classes and functions relevant to constructing t-orders."""

import os

#Pair Classes:

class Pair:
    """Defines a (variant, environment) pair found in factorial typologies."""

    def __init__(self, variant, environment):
        """Create a pair with the given variant and environment."""

        self.pair = (variant, environment)

    def __getVariant(self):
        return self.pair[0]

    def __getEnvironment(self):
        return self.pair[1]

    variant = property(__getVariant)
    environment = property(__getEnvironment)

    def __eq__(self, other):
        if not isinstance(other, Pair):
            return False
        
        return self.pair == other.pair

    def __str__(self):
        return '<'+ str(self.pair[1]) + ', ' + str(self.pair[0]) + '>'

    def __hash__(self):
        return hash(self.pair)

class CompoundPair:
    """Defines a list of (variant, environment) pairs."""

    def __init__(self, pairs):
        self.pairs = pairs

    def __eq__(self, other):
        if not isinstance(other, CompoundPair):
            return False
        
        if len(self.pairs) == len(other.pairs):
            for i in range(len(self.pairs)):
                if self.pairs[i] != other.pairs[i]:
                    return False

            return True

        return False

    def __str__(self):
        output = ''

        for p in self.pairs[:-1]:
            output += (str(p) + '\\n')

        output += str(self.pairs[-1])

        return output

    def contains(self, pair):
        """Determine if pair is contained in this CompoundPair."""

        if not isinstance(pair, Pair):
            return False

        return pair in self.pairs
    

#Factorial Typology File-Reading Functions:

def ParseOutput(words):
    """Create a (variant, environment) pair from a line of text."""

    variant = words[1]
    environment = words[0]

    variant = variant.replace('>','')
    variant = variant.replace('(actual)', '')

    environment = environment.strip()
    variant = variant.strip()

    return Pair(variant, environment.strip('/'))

def ReadDialect(file):
    """Read a dialect from an output set in the given file."""

    words = file.readline().split('-->')

    output = []
    while len(words) > 1:
	output.append(ParseOutput(words))
	words = file.readline().split('-->')

    return output

def ReadVariants(file):
    """Read all the variants from a file."""

    variants = []
    line = file.readline().strip()
    
    while True:
	if len(line) and line[0] == '[':
            word = line.split(':')
            
            variant = word[0]
            variant = variant.strip('[]')

            variant = variant.replace('>','')
            variant = variant.strip()

            if variant not in variants:
                variants.append(variant)

        elif len(line) and line[0:3] == 'IV.':
            break

        line = file.readline().strip()

    return variants

def ReadFactorialTypology(filename):
    """Read in the factorial typology from a given OTSoft output file."""

    dialects = []
    pairs = []
    
    f = open(filename)

    try:
        #Look for output sets.
        line = f.readline()
        while line:
            words = line.split()

            if  len(words) and words[0] == 'III.':
                #Read in all possible variants.
                variants = ReadVariants(f)

            if len(words) and words[0] == 'OUTPUT' and words[1] == 'SET':
                #Skip a line to get to the useful information.
                f.readline()
                output = ReadDialect(f)

                dialects.append(set(output))
                pairs.extend(output)

            line = f.readline()

        f.close()

        #Construct all the environments.
        environments = [p.environment for p in dialects[0]]
    
        return {'dialects' : dialects, 'pairs' : pairs,
                'variants' :  variants, 'environments' : environments}
    
    except:
        #For some reason we couldn't read the file.
        #If this happens with good files, then it's probably a
        #programming error.
        f.close()
        return None


#T-Order Related Functions:

def ComputeTOrder(dialects, pairs):
    """Compute a t-order.

    Compute a t-order from given dialects and pairs, returning a list of
    (Pair, Pair) tuples.
    """

    #Build a list of possible pairings.
    pairings = set()
    for i in pairs:
        for j in pairs:
            if i.environment != j.environment:
                newpair = (i,j)
                pairings.add(newpair)
    
    #Construct the t-order.

    result = []

    for p in pairings:
        implied = True
        
        for d in dialects:
            if p[0] in d and p[1] not in d:
                implied = False
                break

        if implied:
            result.append(p)


    #I'm keeping the old algorithm in here just in case I screw something up.
    #for i in pairs:
	#for j in pairs:
	    #if i.environment != j.environment:
		
		#This pair is valid, so see check for implication.
		#implied = True

		#for d in dialects:
		    #if i in d and j not in d:
			#implied = False
			#break
                
		#if implied:
		    #result.append((i,j))
		    
    #Remove duplicates from result.
    #noduplicates = []
    #for r in result:
	#if r not in noduplicates:
	    #noduplicates.append(r)

    #return noduplicates
    return result

def FindEquivalentNodes(torder):
    """Return a list of sets of equivalent t-order nodes"""
    equivalents = []
    
    for pair in torder:
        reverse = (pair[1], pair[0])
        if reverse in torder and reverse not in equivalents:
            equivalents.append(pair)

    oldlist = [set([e[0],e[1]]) for e in equivalents]
    newlist = []
    
    while True:
        combined = 0

        while len(oldlist):
            pair = oldlist.pop()
            equal = None
            
            for other in oldlist:
                if pair & other:
                    newlist.append(pair | other)
                    combined += 1
                    equal = other
                    break
                
            if equal:
                oldlist.remove(equal)
            else:
                newlist.append(pair)

        if not combined:
            break

        oldlist = newlist
        newlist = []

    return [list(n) for n in newlist]

def CollapseEquivalentNodes(torder):
    """Collapse all equivalent nodes in the given t-order.

    Given a t-order, return a new t-order with all equivalent nodes
    collapse into a single CompoundPair node. Also return a list of
    CompoundPairs that have no edges connecting them to the rest of the
    nodes.
    """

    #Get all equivalent pairs.
    equipairs = FindEquivalentNodes(torder)

    #Create the CompoundPairs.
    comppairs = []
    for e in equipairs:
        comppairs.append(CompoundPair(e))

    #Create a list to determine which CompoundPairs get used in the t-order.
    attached = [False for e in comppairs]

    comprange = range(len(comppairs))

    #Create a new t-order with altered edges.
    newtorder = []
    for implication in torder:
        pair0 = implication[0]
        pair1 = implication[1]

        addimplication = True

        for i in comprange:
            c = comppairs[i]
            
            haspair0 = c.contains(pair0)
            haspair1 = c.contains(pair1)

            if haspair0 and haspair1:
                addimplication = False
                break

            if haspair0:
                pair0 = c
                attached[i] = True
            if haspair1:
                pair1 = c
                attached[i] = True

        if addimplication:
            newtorder.append((pair0, pair1))

    #Create a list of unattached CompoundPairs.
    unused = []

    for i in comprange:
        if not attached[i]:
            unused.append(comppairs[i])

    return (newtorder, unused)

def IsConnected(edge0, edge1):
    """Given two edges (pairs of Pair objects) in a t-order, determine
    if they are connected."""

    if edge0[0] == edge1[0]:
        return True
    if edge0[1] == edge1[0]:
        return True
    if edge0[0] == edge1[1]:
        return True
    if edge0[1] == edge1[1]:
        return True

    return False

def SplitTOrder(torder):
    """Split the given t-order into a list of disjoint smaller t-orders."""

    oldlist = [[pair] for pair in torder]
    newlist = []

    while True:
        combined = 0

        while len(oldlist):
            pairlist = oldlist.pop()
            equal = None
            
            for other in oldlist:

                connected = False
                
                for p in pairlist:
                    for o in other:

                        if IsConnected(p, o):
                            newlist.append(pairlist + other)
                            connected = True
                            combined += 1
                            equal = other
                            break
                        
                    if connected:
                        break

                if connected:
                    break

            if equal:
                oldlist.remove(equal)
            else:
                newlist.append(pairlist)

        if not combined:
            break

        oldlist = newlist
        newlist = []

    return newlist


#File Writing Functions:

def WriteExistingPairs(file, pairs, environments, variants):
    """Write out the existing and non-existing pairs to a file."""
    
    exists = set(pairs)
    possibles = set([Pair(v, e) for v in variants for e in environments])
    possibles = possibles - exists

    file.write(str(len(exists)) + ' Existing Pairs:\n')

    for p in exists:

        file.write(str(p) + '\n')
        
    file.write('\n')
    
    file.write(str(len(possibles)) + ' Non-existing Pairs:\n')

    for p in possibles:
        file.write(str(p) + '\n')

def WriteTOrder(file, torder):
    """Write the given t-order information to a given file."""

    for pair in torder:
	file.write(str(pair[0]) + ' --> ' + str(pair[1]) + '\n')

def WriteDOTGraph(file, torder, extra):
    """Write the given t-order (with extra nodes) to a graphvis dot file."""

    file.write('digraph torder {\n')
    file.write('\tnode [shape=box];\n')
    for pair in torder:
	firstpair = '\"' + str(pair[0]) + '\"'
        secondpair = '\"' + str(pair[1]) + '\"'
	line = '\t' + firstpair + ' -> ' + secondpair + ';\n'

	file.write(line)

    for node in extra:
        file.write('\tnode \"' + str(node) + '\";\n')

    file.write('}')

def OutputAndReduce(torder, extra, filename, format):
    """Given a torder (w/ extra nodes) and a format supported by dot, output
    a reduced graph (from calling tred.exe) into the format."""

    #TODO: Come up with a useful way of outputing any errors.

    #Find dot and tred.
    progs = {'dot': '', 'tred': ''}
    for path in os.environ['PATH'].split(os.pathsep):
        if os.path.exists(path + os.path.sep + 'dot.exe'):
            progs['dot'] = path + os.path.sep + 'dot.exe'
        if os.path.exists(path + os.path.sep + 'tred.exe'):
            progs['tred'] = path + os.path.sep + 'tred.exe'
    if progs['dot'] == '':
        raise OSError('Unable to find dot.exe.')
    if progs['tred'] == '':
        raise OSError('Unable to find tred.exe')
    
    #Create a temporary dot file to call tred on.
    name = 'temp_torder'
    
    temp = open(name, 'w')
    WriteDOTGraph(temp, torder, extra)
    temp.close()

    #Reduce the graph with tred.
    stdin, stdout, stderr = os.popen3(progs['tred'] + ' ' + name, 'b')

    reduced = stdout.read()
    
    #Remove extra lines that tred adds.
    #TODO: Figure out why it does this.
    reduced = reduced.replace('\\\r\n','')

    stdin.close()
    stderr.close()
    stdout.close()

    #Write out the reduced graph to a file.
    rname = 'temp_reducedtorder'
    temp = open(rname, 'w')
    temp.write(reduced)
    temp.close()

    #Finally, create the output file.
    
    stdin, stdout, stderr = os.popen3(progs['dot'] + ' -T' + format
                                      + ' -o' + '\"' + filename + '\" ' + rname, 'b')

    output = stdout.read()
    errors = stderr.read()

    stdin.close()
    stdout.close()
    stderr.close()

    #Remove temporary files.
    os.remove(name)
    os.remove(rname)

def OutputSplitTOrder(torder, extra, filename, format):
    """Given a t-order, with extra nodes, a filename and format, split the
    t-order into connected components and output to different files.
    """

    torders = SplitTOrder(torder)

    filenumber = 0

    try:
        index = filename.index('.')
    except ValueError:
        index = len(filename)

    name = filename[:index]
    extension = filename[index:]

    for t in torders:
        fname = name + str(filenumber) + extension
        OutputAndReduce(t, [], fname, format)
        filenumber += 1

    if extra:
        fname = name + str(filenumber) + extension
        OutputAndReduce([], extra, fname, format)
            
