"""Handles a gui for building t-orders."""

from Tkinter import *
import tkFileDialog, tkMessageBox
import torder, threading

supported = [('DOT', '.canon'), ('DOT', '.dot'), ('DOT', '.xdot'),
             ('Client-Side Imagemap', '.cmap'), ('Dia Format', '.dia'), ('FIG', '.fig'),
             ('GD Format', '.gd'), ('GD2 Format', '.gd2'), ('GIF Image', '.gif'),
             ('HP-GL/2', '.hpgl'), ('Server- and Client-Side Imagemap', '.imap'),
             ('Server-Side and Client-Side Imagemap', '.cmapx'),
             ('Server-Side Imagemap', '.ismap'), ('JPEG Image', '.jpg'),
             ('FrameMaker MIF Format', '.mif'), ('MetaPost', '.mp'), ('PCL', '.pcl'),
             ('PIC', '.pic'), ('Simple Text Format', '.txt'),
             ('Portable Network Graphics Format', '.png'),
             ('PostScript', '.ps'), ('PostScript for PDF', '.ps2'),
             ('Scalable Vector Graphics', '.svg'), ('Scalable Vector Graphics', '.svgz'),
             ('VRML', '.vrml'), ('Visual Thought Format', '.vtx'),
             ('Wireless BitMap Format', '.wbmp')]

extensions = [format[1].lstrip('.') for format in supported]

class TOrderGUI:
    """The main class for the t-order gui."""

    def __init__(self, master):
        self.frame = Frame(master, borderwidth = 3)
        self.frame.grid()

        #Not sure if this is necessary, but I'll keep it here anyways.
        self.datalock = threading.Semaphore()

        self.savepairs = IntVar()
        self.savepairs.set(0)

        self.collapsecycles = IntVar()
        self.collapsecycles.set(0)

        self.splitgraphs = IntVar()
        self.splitgraphs.set(0)

        self.infile = StringVar()
        self.infile.set('None')
        
        self.status = StringVar()
        self.status.set('No Input File')
        
        self.facttype = None
        self.torder = None

        self.createInputWidgets()
        self.createGenWidgets()
        self.createOutputWidgets()

    def createInputWidgets(self):
        #Create the input widgets.
        inlabel = Label(self.frame, text = 'Input File:')
        inlabel.grid(row = 0, sticky = E)

        infilename = Label(self.frame, textvariable = self.infile,
                        width = 50, anchor = W)
        infilename.grid(row = 0, column = 1,
                        columnspan = 2, sticky = W)
        
        inbutton = Button(self.frame, text = 'Open',
                          command = self.getInFile)
        inbutton.grid(row = 0, column = 3, sticky = W)

    def createGenWidgets(self):
        #T-Order Generation widgets.
        
        self.genbutton = Button(self.frame, text = 'Generate T-Order',
                           command = self.generate)    
        self.genbutton.grid(row = 2, columnspan = 3)

        status = Label(self.frame, text = 'Status:')
        status.grid(row = 1, column = 0, sticky = E)

        statlabel = Label(self.frame, textvariable = self.status)
        statlabel.grid(row = 1, column = 1, sticky = W)
        

    def createOutputWidgets(self):
        #T-Order/Graph Saving widgets.

        pairs = Checkbutton(self.frame, text = 'Save Pair Possibilities',
                            variable = self.savepairs)
        pairs.grid(row = 3, column = 1, sticky = W)

        collapse = Checkbutton(self.frame, text = 'Collapse Cycles',
                               variable = self.collapsecycles)
        collapse.grid(row = 4, column = 1, sticky = W)

        split = Checkbutton(self.frame, text = 'Split Disjoint Graphs',
                            variable = self.splitgraphs)
        split.grid(row = 5, column = 1, sticky = W)

        savetorder = Button(self.frame, text = 'Save T-Order',
                            command = self.saveTOrder)
        savetorder.grid(row = 6, column = 1, sticky = W)

        savegraph = Button(self.frame, text = 'Save Graph',
                           command = self.saveGraph)
        savegraph.grid(row = 6, column = 1)
        

    def saveTOrder(self):
        if self.status.get() == 'Complete!':
            prompt = 'Save T-Order File'
            outputfile = tkFileDialog.asksaveasfilename(title = prompt)

            if not outputfile:
                return

            f = open(outputfile, 'w')

            f.write('Input File: ' + self.infile.get() + '\n\n')

            if self.savepairs.get():
                pairs = self.facttype['pairs']
                environments = self.facttype['environments']
                variants = self.facttype['variants']
                torder.WriteExistingPairs(f, pairs, environments, variants)
                f.write('\n')

            f.write('Constructed T-Order:\n')
            torder.WriteTOrder(f, self.torder)
        else:
            tkMessageBox.showerror('No T-Order',
                                   'T-Order Has Not Been Generated.')
            
    def saveGraph(self):
        if self.status.get() == 'Complete!':
            prompt = 'Save Graph File'
            outputfile = tkFileDialog.asksaveasfilename(title = prompt,
                                                        defaultextension = 'jpg',
                                                        filetypes = supported)

            if not outputfile:
                return

            nameparts = outputfile.split('.')
            ext = nameparts[len(nameparts) - 1]
            format = 'jpg'

            if ext == 'txt':
                format = 'plain'
            elif ext in extensions:
                format = ext

            finaltorder = self.torder
            extras = []
            
            if self.collapsecycles.get():
                data = torder.CollapseEquivalentNodes(self.torder)
                finaltorder = data[0]
                extras = data[1]

            try:
                if not self.splitgraphs.get():
                    torder.OutputAndReduce(finaltorder, extras, outputfile,
                                           format)
                else:
                    torder.OutputSplitTOrder(finaltorder, extras, outputfile,
                                             format)
            except OSError:
                tkMessageBox.showerror('Missing Programs',
                                       'Cannot find necessary Graphvis\
                                       programs!')
            
        else:
            tkMessageBox.showerror('No T-Order',
                                   'T-Order Has Not Been Generated.')

    def getInFile(self):
        if self.status.get() == 'Generating...':
            tkMessageBox.showerror('Please Wait',
                                   'The t-order is being generated. Please wait.')
            return
        
        self.datalock.acquire()
        
        prompt = 'Open Factorial Typology File'
        oldname = self.infile.get()
        
        self.infile.set(tkFileDialog.askopenfilename(title = prompt))

        filename = self.infile.get()

        if filename and filename != oldname:
            self.facttype = torder.ReadFactorialTypology(self.infile.get())
            if not self.facttype:
                self.infile.set(oldname)
                tkMessageBox.showerror('Open File',
                                       'Invalid Factorial Typology File.')
            else:
                self.torder = None
                self.status.set('Ready')
                self.genbutton.config(state = NORMAL)
        else:
            self.infile.set(oldname)

        self.datalock.release()
        

    def generate(self):
        if self.facttype and not self.torder:
            dialects = self.facttype['dialects']
            pairs = self.facttype['pairs']

            self.status.set('Generating...')
            self.genbutton.config(state = DISABLED)

            #Create a new thread to generate the torder in.
            genthread = threading.Thread(target = self.buildTOrder,
                                         args = (dialects, pairs))
            
            genthread.start()
            

    def buildTOrder(self, dialects, pairs):
        self.datalock.acquire()
        
        self.torder = torder.ComputeTOrder(dialects, pairs)
        self.status.set('Complete!')

        self.datalock.release()
        
            
root = Tk()
root.title('T-Order Generator')
root.resizable(False, False)

app = TOrderGUI(root)

import profile
root.mainloop()
